# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Public API for tf.keras.__internal__.backend namespace.
"""

import sys as _sys

from keras.backend import _initialize_variables as initialize_variables
from keras.backend import track_variable