# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Public API for tf.keras.datasets.imdb namespace.
"""

import sys as _sys

from keras.datasets.imdb import get_word_index
from keras.datasets.imdb import load_data